"""
    This behavior file contains the logic for a subset of commands. Logic specific to
    commands should be implamented in corresponding behavior files.
"""

from pathlib import Path
from cli import pyke
import requests
import click
import time
import json
import sys
import os

class WidgetVersionBehavior:
    def __init__(self, profile=None, widget=None, format=None, filter=None, page=None, pagesize=None, json=False):
        self.util = pyke.Util(profile=profile)
        if profile is not None:
            self.context = pyke.auth.load_cache()
        self.object_type = 'widget'
        self.json = json
        self.profile = profile
        self.widget = widget
        self.format = format
        self.filter = filter
        self.page = page
        self.pagesize = pagesize

    def list(self, widget):
        widget_id = self.util.lookup_object_id(self.object_type, widget)
        data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}

        handlers = {
          'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
        }

        resp = self.util.cli_request('GET',
            self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions?page={page}&pagesize={pagesize}&{filter}',
            {'widget_id': widget_id, **data}))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table(resp['payload']['data'], self.format, handlers=handlers)
        self.util.print_record_count(resp)

    def start_stop(self, action, widget, version):
        widget_id = self.util.lookup_object_id(self.object_type, widget)
        version_data = None

        try:
            version_id = int(version)
        except:
            major, minor, patch = self.util.parse_version_number(version)

            version_data = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions?major={major}&minor={minor}&patch={patch}',\
                { 'widget_id': widget_id, 'major': major, 'minor': minor, 'patch': patch }))['payload']['data']

            if version_data is None or len(version_data) < 1:
                raise click.ClickException(click.style("Version not found", fg='red'))
            else:
                version_id = version_data[0].get('id')
                current_state = version_data[0].get('actualState')

        if version_data is None:
            version_data = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}',\
                {'widget_id': widget_id, 'version_id': version_id} ))['payload']['data']

            if version_data is None or len(version_data) < 1:
                raise click.ClickException(click.style("Version not found", fg='red'))

            current_state = version_data.get('actualState')

        if current_state not in ['stopped', 'running', 'error']:
            raise click.ClickException(click.style("To start or stop a version it must have a 'stopped', 'running' or 'error' state. The current state of this version is {}. \
                A version goes through several states after being uploaded to Lumavate. It must finish this process before this command can be run. \
                To check its current state, run 'luma widget-version ls'.".format(current_state), fg='yellow'))

        if action == 'stop' and current_state != 'running':
            raise click.ClickException(click.style("This version is not running", fg='yellow'))
        if action == 'start' and current_state == 'running':
            raise click.ClickException(click.style("This version is already running", fg='yellow'))

        titles = {
            'id': 'id',
            'actualState': 'Current State',
            'versionNumber': 'Version #',
            'createdAt': 'Created At',
            'createdBy': 'Created By'
        }

        data = {
            "action": action,
            "containerId": widget_id,
            "id": version_id
        }

        resp = self.util.cli_request('POST', self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}/{action}',\
            {'widget_id': widget_id, 'version_id': version_id, 'action': action}), data=data)

        if self.json:
            click.echo(json.dumps(resp))
            return

        status_id = resp['payload']['data']['results']['statusId']
        progress = 0
        error = None
        if action == 'start':
            bar_name = "Starting widget"
        else:
            bar_name = "Stopping widget"

        with click.progressbar(length=100, label=bar_name) as bar:
            while progress < 100 and error is None:
                status_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))
                progress = status_resp['payload']['data']['overallPercent']
                error = status_resp['payload']['data']['errorMessage']
                bar.update(progress)

        click.echo(' ')
        ver_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}', {'widget_id': widget_id, 'version_id': version_id}))

        handlers = {
          'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
        }

        self.util.print_table([ver_resp['payload']['data']], self.format, handlers=handlers, titles=titles)

    def run(self, widget, version, command, target):
        widget_id = self.util.lookup_object_id(self.object_type, widget)
        version_id = self.util.get_version_id(self.object_type, widget_id, version)

        data = {
            "command": command,
            "target": target
        }

        resp = self.util.cli_request('POST', self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}/exec',\
            {'widget_id': widget_id, 'version_id': version_id}), json=data)

        if self.json:
            click.echo(json.dumps(resp))
            return

        status_id = resp['payload']['data']['statusId']
        progress = 0
        error = None
        with click.progressbar(length=100, label='Executing command') as bar:
            while progress < 100:
                status_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))
                progress = status_resp['payload']['data']['overallPercent']
                bar.update(progress)
                if status_resp['payload']['data']['errorMessage'] is not None:
                    click.echo('There was an error running the command')
                    raise click.ClickException('Status: {}'.format(status_resp['payload']['data']))

        click.echo(' ')
        click.echo('Command Output: ')
        click.echo(' ')
        try:
            status_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))['payload']['data']
            if target == 'one':
                summary = status_resp['summary']['results']
                if 'output' in summary[0].keys():
                    if isinstance(summary[0]['output'], list):
                        for x in summary[0]['output']:
                            click.echo(x)
                    else:
                        click.echo(summary)
            else:
                summary = status_resp['summary']['results']
                for x in summary:
                    click.echo(x)

        except:
            click.echo(status_resp)

    def logs(self, widget, version):
        widget_id = self.util.lookup_object_id(self.object_type, widget)
        version_id = self.util.get_version_id(self.object_type, widget_id, version)

        resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}/logs',\
            {'widget_id': widget_id, 'version_id': version_id}))

        if self.json:
            click.echo(json.dumps(resp))
            return

        for x in resp['payload']['data']:
            click.echo(x)

    def create(self, widget, port, widget_file_path, docker_image,\
        from_version, version_number, version_bump, env, label, format):
        widget_id = self.util.lookup_object_id(self.object_type, widget)

        if docker_image is not None and docker_image != '':
            if not self.json:
                click.echo("Getting docker image ")
            cu = pyke.ContainerUtil()
            # save_image() returns the file path of the zipped docker image
            widget_file_path = cu.save_image(docker_image)

        if from_version != '' and from_version is not None:
            try:
                from_version_id = int(from_version)
                version = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{from_version_id}',\
                    {'widget_id': widget_id, 'from_version_id': from_version_id} ))['payload']['data']
            except:
                version = self.util.get_version(self.object_type, widget_id, from_version)

            if version is None:
                raise click.BadParameter(click.style("Widget version not found", fg='red'), param_hint="--version-number")

            if version_number is not None and version_number != '':
                major, minor, patch = self.util.parse_version_number(version_number)

                version['major'] = major
                version['minor'] = minor
                version['patch'] = patch

            else:
                version[version_bump] = version[version_bump] + 1
                if version_bump in ['major', 'minor']:
                    version['patch'] = 0
                if version_bump == 'major':
                    version['minor'] = 0

        else:
            version = {}
            if label is None or label == '':
                label = input("Label: ")
                if label is None or label == '':
                    raise click.BadParameter(click.style("If you don't create from an old version then you must provide a --label", fg='red'), param_hint="--label")

            if version_number is None or version_number == '':
                version_number = input("Version Number: ")
                major, minor, patch = self.util.parse_version_number(version_number)

                version['major'] = major
                version['minor'] = minor
                version['patch'] = patch

                if version_number is None or version_number == '':
                    raise click.BadParameter(click.style("If you don't create from an old version then you must provide a --version-number in the format\
                        '<major: int>.<minor: int>.<patch: int>' ", fg='red'), param_hint="--version-number")

            if widget_file_path is None or widget_file_path == '':
                widget_file_path = input("Widget File Path: ").replace(" ", "")
                if widget_file_path is None or widget_file_path == '':
                    raise click.ClickException(click.style("If you don't create from an old version then you must provide --widget-file-path or --docker-image", fg='red'))

            if port is None or port == '':
                port = int(input("Port: "))
                if port is None or port == '':
                    raise click.ClickException(click.style("If you don't create from an old version then you must provide a --port for your widget", fg='red'))

            else:
                major, minor, patch = self.util.parse_version_number(version_number)

                version['major'] = major
                version['minor'] = minor
                version['patch'] = patch

        if widget_file_path is not None and widget_file_path != '':
            if not self.json:
                click.echo("Uploading image to Lumavate")
                click.echo('Image Size: {}'.format(self.util.get_size(widget_file_path)))

            version['tag'] = None
            version['ephemeralKey'] = self.util.upload_ephemeral(widget_file_path, 'application/gz')

        version['platformVersion'] = 'v2'
        if 'env' not in version.keys():
            version['env'] = {}

        for var in env:
            try:
                env_dict = json.loads(var)
                version['env'] = { **version['env'], **env_dict}
            except Exception as e:
                print("Could not serialize {}".format(var))
                print(e)

        if label != '' and label is not None:
            version['label'] = label

        if port != '' and port is not None:
            version['port'] = port

        version['instanceCount'] = 1

        resp = self.util.cli_request('POST',
            self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions', {'widget_id': widget_id}), data=json.dumps(version))

        if self.json:
            click.echo(json.dumps(resp))
            return

        # TO DO: Add progress bar after create for loading/validating widget
        progress = 0
        status_id = None
        error = None

        if 'results' in resp['payload']['data'].keys():
            status_id = resp['payload']['data']['results']['statusId']
        if status_id:
            with click.progressbar(length=100, label='Creating widget') as bar:
                while progress < 100 and error is None:
                    status_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))
                    progress = status_resp['payload']['data']['overallPercent']
                    error = status_resp['payload']['data']['errorMessage']
                    bar.update(progress)


        handlers = {
            'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
        }

        self.util.print_table([resp['payload']['data']], format, handlers=handlers)

        if docker_image is not None and docker_image != '':
            try:
                os.remove(widget_file_path)
            except:
                click.echo('Failed to pick up our mess...')

    def update(self, widget, version, label, format):
        widget_id = self.util.lookup_object_id(self.object_type, widget)
        version_id = self.util.get_version_id(self.object_type, widget_id, version)

        post_data = {
          'widgetId': widget_id,
          'id': version_id,
          'label': label
        }

        resp = self.util.cli_request('PUT',\
            self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}',
            {'widget_id': widget_id, 'version_id': version_id}), data=json.dumps(post_data))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table([resp['payload']['data']], format)

    def delete(self, widget, version, version_mask, format):
        if version is None and version_mask is None:
            version = str(input("Version: "))
            if version is None or version == '':
                raise click.BadParameter(click.style('You must provide either a --version or a --version-mask', fg='red'))

        widget_id = self.util.lookup_object_id(self.object_type, widget)

        if version is not None:
            version_id = self.util.get_version_id(self.object_type, widget_id, version)

            resp = self.util.cli_request('DELETE',\
                self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}', {'widget_id': widget_id, 'version_id': version_id}))

            if self.json:
                click.echo(json.dumps(resp))
                return

            self.util.print_table([resp['payload']['data']], format)

        else:
            versions = self.util.get_versions_from_mask(self.object_type, widget_id, version_mask)

            if self.json:
                resp_list = {}
                resp_list['responses'] = []
                for v in versions:
                    resp = self.util.cli_request('DELETE',
                        self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}', {'widget_id': widget_id, 'version_id': v['id']}))
                    resp_list['responses'].append(resp)
                click.echo(json.dumps(resp_list))
                return

            with click.progressbar(versions, label='Deleting Versions') as bar:
                for v in bar:
                    resp = self.util.cli_request('DELETE',
                        self.util.build_url('{app}/iot/v1/containers/{widget_id}/versions/{version_id}', {'widget_id': widget_id, 'version_id': v['id']}))

            self.util.print_table(versions, format)



"""
    The auth class is resposible for storing, reading and updating tokens and config settings.
    The CLI should cache tokens until a new one is required. If that's the case the auth class
    retrieves the config variables so a new token can be requested from the platform. Upon success
    the new token is cached and used for subsequent commandsself.
"""

from functools import wraps
from pathlib import Path
from pyparsing import *
import dateutil.parser
import requests
import click
import copy
import json
import os


class auth:

    def get_profile(config_data, name):
        return config_data['profiles'].get(name)

    def get_env(config_data, name):
        return config_data['envs'].get(name)

    def set_cache(data):
        clean_data = auth.sanatize_data(data)
        with open(auth.get_cache_file(), 'w+') as outfile:
            json.dump(data, outfile, indent=4, sort_keys=True)

    def clear_cache():
        with open(auth.get_cache_file(), 'w+') as outfile:
            json.dump({}, outfile, indent=4, sort_keys=True)

    def set_cache_token(token):
        cache = auth.load_cache()
        cache['accessToken'] = token
        auth.set_cache(cache)

        return cache

    def get_profile_names(ctx, args, incomplete):
        config = auth.load_config()
        names = config['profiles'].keys()
        return [k for k in names if incomplete in k]

    """
        Data coming from the config file and cache sometimes have extra whitespace 
        appended to the values. Their are casses where, values being read with
        extra whitespace are then being written back to the config.
        Track this down and prevent this from happening.
    """
    def sanatize_data(data):
        return { k:v.replace(" ", "") if isinstance(v, str) and k != 'orgName' else v for k,v in data.items() }

    def set_profile_token(profile, token):
        config_data = auth.load_config()
        config_data['profiles'][profile]['accessToken'] = token
        auth.save_config(config_data)

        return config_data

    def get_cache_file():
        cache_path = str(Path.home()) + '/.luma_cache'
        if not os.path.exists(cache_path):
            with open(cache_path, 'w+') as cache:
                json.dump({}, cache)

        return cache_path

    def load_cache():
        with open(auth.get_cache_file()) as json_file:
            data = json.load(json_file)

        return data

    def save_config(data):
        with open(auth.get_config_file(), 'w+') as outfile:
            json.dump(data, outfile, indent=4, sort_keys=True)

    def load_config():
        try:
            with open(auth.get_config_file()) as json_file:
                data = json.load(json_file)
            return data
        except Exception as e:
            print('No config file or bad config file found. Making a new one.')
            auth.save_config({ "envs": {}, "profiles": {} })

            return {
                "envs": {},
                "profiles": {}
            }

    def get_config_file():
        config_path = str(Path.home()) + '/.luma_cli_config'
        if not os.path.exists(config_path):
            with open(config_path, 'w') as config:
                json.dump({ "envs": {}, "profiles": {} }, config)

        return config_path

    """
        This decorator should be used on any command that requires company context.
        It ensures that the user runs the command with a valid profile and that the
        profile is able to obtain a token with the correct company context.
    """
    def profile_context_required(f):
        @wraps(f)
        def decorated_func(*args, **kwargs):
            try:
                profile = kwargs.get('profile')
                config_data = auth.load_config()
                profile_data = auth.get_profile(config_data, profile)

                if profile_data is None:
                    raise click.BadParameter(click.style('You must provide the name of a profile that you have configured', fg='red'))
                if profile_data.get('orgId') is None or profile_data.get('orgId') == '':
                    raise click.BadParameter(click.style('You must finish configuring this profile before you can use it', fg='red'))

                clean_data = auth.sanatize_data(profile_data)

                clean_data['profileName'] = profile
                env_data = auth.sanatize_data(auth.get_env(config_data, clean_data.get('env')))

                cache_data = {**clean_data, **env_data}

                # Protect against error state where 'accessToken' would become null
                if cache_data.get("accessToken") is None:
                    cache_data['accessToken'] = " "

                auth.set_cache(cache_data)

            except Exception as e:
                click.echo(click.style(str(e), fg='red'))
                raise click.Abort()

            f(*args, **kwargs)
            auth.clear_cache()

            return
        return decorated_func



"""
    The Util class is comprised of common code that is used throughout the CLI.
    Refreshing tokens, making calls to the platform, building urls, and looking up
    various lumavate objects are few things implamented here.

    TO DO: This needs to be broken up further. Refreshing tokens and all other things
    auth should be broken out into the auth class. Request handling and the various validation
    of objects sprinkled in this code and the behavior files are two other major chunks of code
    that need a new home.
"""

from pathlib import Path
from pyparsing import *
import dateutil.parser
from cli import pyke
import requests
import platform
import urllib
import click
import copy
import json
import sys
import os


ESC = Literal('\x1b')
integer = Word(nums)
escapeSeq = Combine(ESC + '[' + Optional(delimitedList(integer,';')) +
                oneOf(list(alphas)))

uncolor = lambda s : Suppress(escapeSeq).transformString(s)

class Util:
    def __init__(self, context=None, profile=None):
        self.context = pyke.auth.load_cache()
        if profile is None:
            self.profile = self.context.get('profileName')
        else:
            self.profile = profile

    def login_token(self, renew=False):
        if renew:
            self.context['accessToken'] = self.login()['access_token']
            return self.context['accessToken']

        if self.context is not None:
            return self.context.get('accessToken')

    def login_without_context(self, env, config_data=None):
        if config_data is None:
            config_data = pyke.auth.load_config()
        else:
            config_data = config_data

        env = config_data['envs'][env]

        data = {
          'grant_type': 'client_credentials',
          'audience': env['audience'],
          'client_id': env['clientId'],
          'client_secret': env['clientSecret']
        }

        req_data = pyke.auth.sanatize_data(data)

        token_url = env['token'].replace(" ", "")
        resp = self.cli_request('POST', 'https://' + token_url+ '/oauth/token', json=req_data, external=True)

        if resp.status_code != 200:
            click.echo(json.dumps(env))
            click.echo(' ')

            raise click.ClickException(click.style("Env not configured correctly. Make sure your Client ID and Client Secret are correct.", fg='red'))

        return resp.json()['access_token']

    def login(self):
        data = {
          'grant_type': 'client_credentials',
          'audience': self.context['audience'].replace(" ", ""),
          'client_id': self.context['clientId'].replace(" ", ""),
          'client_secret': self.context['clientSecret'].replace(" ", "")
        }

        res = self.cli_request('POST', 'https://{}/oauth/token'.format(self.context['token']), json=data, external=True)

        if res.status_code != 200:
            raise click.ClickException(click.style("Env not configured correctly. Make sure your Client ID and Client Secret are correct.", fg='red'))

        resp = res.json()

        if self.context.get('orgId') is not None:
            url = '{}/auth/v1/session'.format(self.context.get('app'))
            data = {'companyId': self.context.get('orgId')}
            r = requests.post(url, json=data, headers={'Authorization': 'Bearer ' + resp['access_token']})
            if r.status_code != 200:
                raise click.BadParameter(click.style("Not authorized for company id {}".format(self.context.get('orgId')), fg='red'))

        pyke.auth.set_cache_token(resp.get('access_token'))
        pyke.auth.set_profile_token(self.context.get('profileName'), resp.get('access_token'))

        return resp

    def print_record_count(self, data):
        d = data['payload']
        if d['totalItems'] > d['currentItemCount']:
            d['startItem'] = (d['page'] - 1) * d['pageSize'] + 1
            d['endItem'] = d['startItem'] - 1 + d['currentItemCount']
            count_str = 'showing {startItem} - {endItem} of {totalItems} total items (page {page} / {totalPages})'
            click.echo(click.style(count_str.format(**d), fg='yellow'))

    def load_users(self):
        resp = self.cli_request('GET', self.build_url('{app}/auth/v1/users'))

        users = {}
        for u in resp['payload']['data']:
            users[u['id']] = u['emailAddress']

        return users

    def print_table(self, data, format, titles={}, handlers={}):
        headers = {}
        stats = {}
        users = {}

        if 'createdBy' in format or 'lastModifiedBy' in format:
            users = self.load_users()

        for c in data:
            child_records = {}
            for k in c:
                if isinstance(c[k], dict):
                    for sk in c[k]:
                        child_records[k + '__' + sk] = c[k][sk]

            c.update(child_records)

        if len(data) > 0:
            for k in data[0]:
                headers[k] = titles.get(k, k)
        else:
            click.echo('No Records Found')
            return

        for k in headers:
            stats[k] = len(uncolor(headers[k]))

        # TO DO: Cache users and reload when we can't find one
        for c in data:
            for k in c:
                if k in ['createdAt', 'lastModifiedAt']:
                    cd = dateutil.parser.parse(c[k])
                    c[k] = cd.strftime('%x %X')
                if k in ['createdBy', 'lastModifiedBy']:
                    c[k] = users.get(c[k], 'Not Found ({})'.format(c[k]))
                if stats.get(k, 0) < len(str(uncolor(handlers[k](c[k])) if k in handlers else c[k])):
                    stats[k] = len(str(uncolor(handlers[k](c[k])) if k in handlers else c[k]))

        for k in headers:
            headers[k] = titles.get(k, k).ljust(stats[k])

        for c in data:
            for k in c:
                c[k] = str(handlers[k](c[k]) if k in handlers else c[k])
                c[k] = c[k] + ' ' * (stats[k] - len(uncolor(c[k])))

        click.echo(format.format(**headers))

        for c in data:
            print(format.format(**c))

    def check_response(self, resp):
        if resp.status_code == 500:
            raise click.ClickException(str(click.style('Unhandled Exception: ', fg='red')) + str(resp.text))
        elif resp.status_code == 404:
            raise click.ClickException(click.style('Not Found', fg='red'))
        elif resp.status_code == 400:
            message = resp.json()['payload']['error']
            if 'apiField' in resp.json()['payload']:
                message += ' (' + resp.json()['payload']['apiField'] + ')'
            raise click.ClickException(click.style(message, fg='red'))

        return resp

    def get_headers(self):
        return {
            'Authorization': 'Bearer ' + self.login_token(),
            'Content-Type': 'application/json'
        }

    def build_url(self, format, extra_data = {}):
        data = pyke.auth.sanatize_data(self.context)
        env = copy.deepcopy(data)
        env.update(self.context)
        env.update(extra_data)
        return format.format(**env).replace(" ", "+")

    def cli_request(self, method, url, data=None, json=None, headers=None, params=None, external=False):
        req_args = self.build_req_args(data, json, headers, params, external)
        req = self.get_request_func(method)
        url = url.replace(" ", "")
        if data is not None:
            resp = self.check_response(req(url, data=data, **req_args))
        else:
            resp = self.check_response(req(url, **req_args))

        if resp.status_code != 200 and external is False:
            self.login_token(renew=True)
            req_args = self.build_req_args(data, json, headers, params, external)
            resp = self.check_response(req(url, **req_args))
            if resp.status_code != 200:
                raise click.BadParameter(click.style('Invalid Credentials', fg='red'))

        if not external:
            return resp.json()
        else:
            return resp

    def build_req_args(self, data, json_data, headers, params, external=False):
        if external is False:
            req_headers = self.get_headers()
            if headers is not None:
                req_headers.update(headers)
        elif headers is not None:
            req_headers = headers
        else:
            req_headers = None

        req_args = {}
        if json_data is not None:
            req_args['json'] = json_data
        if req_headers is not None:
            req_args['headers'] = req_headers
        if params is not None:
            req_args['params'] = params

        return req_args

    def get_request_func(self, method):
        method = method.lower()
        methods = {
            'get': requests.get,
            'head': requests.head,
            'post': requests.post,
            'put': requests.put,
            'patch': requests.patch,
            'delete': requests.delete
        }

        return methods[method]

    def list_companies(self, env, print_table=True):
        conf = pyke.auth.load_config()

        handlers = {
            'instanceType': lambda x: 'dev' if x == 'cc' else x,
        }

        headers = {
            'Authorization': 'Bearer ' + self.login_without_context(env),
            'Content-Type': 'application/json'
        }

        app_url = conf['envs'][env]['app'].replace(" ", "")
        res = requests.get('{}/auth/v1/me/companies?sort=name'.format(app_url), headers=headers)
        resp = res.json()['payload']['data']

        titles = {
        'id': 'id',
        'name': 'Org Name',
        'instanceType': 'Org Type',
        'isTest': 'Test Org'
        }

        if print_table is True:
            self.print_table(resp, '{id} {name} {instanceType} {isTest}', handlers=handlers, titles=titles)

        return resp

    def get_company_id_from_resp(self, id, resp):
        try:
            company_id = next(int(rec.get('id')) for rec in resp if rec.get('name').lower() == id.lower())
            return company_id
        except:
            raise click.ClickException(click.style("Company not found. You can only share/unshare with companies you have provisioned.", fg='red'))

    def get_company_id(self, name_or_id, resp):
        try:
            company_id = int(name_or_id)
            return company_id
        except:
            return self.get_company_id_from_resp(name_or_id, resp)

    def get_company_name(self, id, resp):
        return next(rec.get('name') for rec in resp if int(rec.get('id')) == id)

    def upload_ephemeral(self, filename, content_type):
        resp = self.cli_request('GET', self.build_url('{app}/iot/v1/files/ephemeral?contentType=' + urllib.parse.quote(content_type)))
        if '/' in filename:
            file_name = filename.rsplit('/', 1)[-1]
        else:
            file_name = filename

        ephemeral_url = resp['payload']['data']['url']
        ephemeral_data = resp['payload']['data']['fields']
        ephemeral_data['key'] = ephemeral_data['key'].replace('${filename}', file_name)
        ephemeral_data['Content-Type'] = content_type

        post_res = requests.post(
            ephemeral_url,
            data=ephemeral_data,
            files={'file': open(filename, 'rb')})

        return ephemeral_data['key']

    def get_current_access_records(self, object, object_id):
        path = '{}s/{}/access'.format(object, object_id)
        resp = self.cli_request('GET', self.build_url('{app}/iot/v1/' + path))
        return resp['payload']['data']

    def lookup_object_id(self, object_type, id, abort_if_null=True):
        if object_type == 'component-set':
            try:
                object_filter = int(id)
                filter_type = 'id'
            except:
                object_filter = id
                filter_type = 'urlRef'

            resp = self.cli_request('GET',
                self.build_url('{app}/iot/v1/component-sets?{filter_type}={object_filter}',
                    extra_data={'filter_type': filter_type, 'object_filter': object_filter}))


            if len(resp.get('payload').get('data', [])) > 0:
                return resp['payload']['data'][0]['id']

            if abort_if_null:
                click.echo(click.style('{} not found'.format(object_type), fg='red'))
                sys.exit()

        else:
            try:
                object_filter = int(id)
                filter_type = 'id'
            except:
                object_filter = id
                filter_type = 'urlRef'

            resp = self.cli_request('GET',
                self.build_url('{app}/iot/v1/containers?{filter_type}={object_filter}',
                    extra_data={'filter_type': filter_type, 'object_filter': object_filter}))

            if len(resp.get('payload').get('data', [])) > 0:
                return resp['payload']['data'][0]['id']

            if abort_if_null:
                click.echo(click.style('{} not found'.format(object_type), fg='red'))
                sys.exit()

    def save_config(self, data):
        with open(self.get_config_file(), 'w+') as outfile:
            json.dump(data, outfile, indent=4, sort_keys=True)

    def get_size(self, path, image=True):
        stats = os.stat(path)
        byte_size = stats.st_size
        if byte_size < 1048576:
            return '{} KB'.format(round(byte_size / 1024, 2))
        elif byte_size < 1073741824:
            return '{} MB'.format(round(byte_size / 1048576, 2))
        else:
            if image is True:
                click.echo('Hey! Thanks for being an awesome Lumavate customer.')
                click.echo('While you are waiting for this image to upload to the platform you should check this out!')
                click.echo('https://hackernoon.com/tips-to-reduce-docker-image-sizes-876095da3b34')
            return '{} GB'.format(round(byte_size / 1073741824, 2))

    def parse_version_number(self, version_number):
        try:
            major, minor, patch = version_number.split('.')
            for x in [major, minor, patch]:
                if x is None:
                    raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))
                if int(x) < 0:
                    raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))

            return major, minor, patch
        except:
            raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))

    def get_config_file(self):
        config_path = str(Path.home()) + '/.luma_cli_config'
        if not os.path.exists(config_path):
            with open(config_path, 'w') as config:
                json.dump({ "envs": {}, "profiles": {} }, config)

        return config_path

    def get_version_id(self, object_type, parent_id, version):
        try:
            version_id = int(version)
        except:
            _version = self.get_version_from_version_number(object_type, parent_id, version)
            if _version is not None:
                version_id = _version.get('id')

        return version_id

    def get_versions_from_mask(self, object_type, parent_id, version):
        if object_type == 'component-set':
            versions = self.cli_request('GET', self.build_url('{app}/iot/v1/component-sets/{parent_id}/versions',\
                {'parent_id': parent_id}))['payload']['data']
        else:
            versions = self.cli_request('GET', self.build_url('{app}/iot/v1/containers/{parent_id}/versions',\
                {'parent_id': parent_id}))['payload']['data']

        major, minor, patch = version.split('.')

        return sorted([x for x in versions if
            (str(x['major']) == major or major == '*') and
            (str(x['minor']) == minor or minor == '*') and
            (str(x['patch']) == patch or patch == '*')], key=lambda x: (x['major'], x['minor'], x['patch']), reverse=True)

    def get_version_from_version_number(self, object_type, parent_id, version_number):
        major, minor, patch = version_number.split('.')

        if object_type == 'component-set':
            version = self.cli_request('GET', self.build_url('{app}/iot/v1/component-sets/{parent_id}/versions?major={major}&minor={minor}&patch={patch}',\
                { 'parent_id': parent_id, 'major': major, 'minor': minor, 'patch': patch }))['payload']['data']
        else:
            version = self.cli_request('GET', self.build_url('{app}/iot/v1/containers/{parent_id}/versions?major={major}&minor={minor}&patch={patch}',\
                { 'parent_id': parent_id, 'major': major, 'minor': minor, 'patch': patch }))['payload']['data']

        if version is not None and len(version) > 0:
            return version[0]

        else:
            raise click.ClickException(click.style('Version Not Found', fg='red'))

    def get_version(self, object_type, parent_id, version):
        versions = self.get_versions_from_mask(object_type, parent_id, version)
        if versions:
            return versions[0]

        return None

    def resolve_default_icon_path(self):
        pwd = os.path.dirname(os.path.realpath(__file__))
        if platform.system() == 'Windows':
            icon_path = pwd.rsplit('\\', 1)[0]
            return '{}\\images\\default-icon.svg'.format(icon_path)
        else:
            icon_path = pwd.rsplit('/', 1)[0]
            return '{}/images/default-icon.svg'.format(icon_path)



"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

from cli import pyke, behavior
import click

@click.group('widget-version', help="A widget-version is the docker image of widget, \
along with the details it needs to run such as, port number and Env variables.")
def widget_version():
    pass

@click.command('ls')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command', autocompletion=pyke.auth.get_profile_names)
@click.option('--widget', '-w', prompt=True)
@click.option('--format', '-f', default='{id} {actualState} {versionNumber} {label} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--filter', default='')
@click.option('--page', default=1)
@click.option('--pagesize', default=100)
@click.option('--json/--table', default=False)
@pyke.auth.profile_context_required
def list(widget, profile, format, json, **kwargs):
    behavior.WidgetVersionBehavior(profile=profile, format=format, json=json, **kwargs).list(widget)

@click.command('start')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command', autocompletion=pyke.auth.get_profile_names)
@click.option('--widget', '-w', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--format', '-f', default='{id} {actualState} {versionNumber} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False)
@pyke.auth.profile_context_required
def start(widget, version, profile, format, json):
    behavior.WidgetVersionBehavior(profile=profile, format=format, json=json).start_stop("start", widget, version)

@click.command('stop')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command', autocompletion=pyke.auth.get_profile_names)
@click.option('--widget', '-w', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--format', '-f', default='{id} {actualState} {versionNumber} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False)
@pyke.auth.profile_context_required
def stop(widget, version, profile, format, json):
    behavior.WidgetVersionBehavior(profile=profile, format=format, json=json).start_stop("stop", widget, version)

@click.command('exec', help='Pass in a command to be run directly on the docker container.')
@click.argument('command')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command', autocompletion=pyke.auth.get_profile_names)
@click.option('--widget', '-w', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--target', default='one', help="Run this command on ONE replica of the service or ALL replicas.")
@click.option('--json/--table', default=False)
@pyke.auth.profile_context_required
def run(command, widget, version, profile, target, json):
    behavior.WidgetVersionBehavior(profile=profile, json=json).run(widget, version, command, target)

@click.command('logs')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command', autocompletion=pyke.auth.get_profile_names)
@click.option('--widget', '-w', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--json/--table', default=False)
@pyke.auth.profile_context_required
def logs(widget, version, profile, json):
    behavior.WidgetVersionBehavior(profile=profile, json=json).logs(widget, version)

@click.command('add')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command', autocompletion=pyke.auth.get_profile_names)
@click.option('--port', type=int, help='The widgets port number')
@click.option('--widget', '-w', prompt=True)
@click.option('--widget-file-path', '-path', type=click.Path(), help='The image must be a gzipped tar file. Ex: {file_name}.tar.gz')
@click.option('--docker-image', '-image', type=str, help='The name (including the tag) of a docker image from the docker daemon. You must have docker running locally to use this option.')
@click.option('--from-version', '-fv', type=str, help="Take all values from an exsisting version and create a new version. \
Increment the major, minor or patch version by 1 with --major, --minor, --patch flags. Override any old values by passing them in as options.")
@click.option('--version', '-v', type=str, default=None, help="The version number of the new widget version")
@click.option('--patch', 'version_bump', flag_value='patch', default=True)
@click.option('--minor', 'version_bump', flag_value='minor')
@click.option('--major', 'version_bump', flag_value='major')
@click.option('--env-var', 'env', default=None, multiple=True, help="The environment variables to add to docker image when it is run. It must be valid json wrapped in single quotes.")
@click.option('--label', '-l', type=click.Choice(['prod', 'dev', 'old']))
@click.option('--format', '-f', default='{id} {actualState} {versionNumber} {label} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False)
@pyke.auth.profile_context_required
def create(widget, profile, port, widget_file_path,\
    docker_image, from_version, version, version_bump, env, label, format, json):
    behavior.WidgetVersionBehavior(profile=profile, json=json).create(widget, port,\
        widget_file_path, docker_image, from_version, version, version_bump, env, label, format)

@click.command()
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command', autocompletion=pyke.auth.get_profile_names)
@click.option('--widget', '-w', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--label', '-l', prompt=True)
@click.option('--format', '-f', default='{id} {versionNumber} {label} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False)
@pyke.auth.profile_context_required
def update(widget, version, label, profile, format, json):
    behavior.WidgetVersionBehavior(profile=profile, json=json).update(widget, version, label, format)

@click.command('rm')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command', autocompletion=pyke.auth.get_profile_names)
@click.option('--widget', '-w', prompt=True)
@click.option('--version-mask', '-vm', type=str, help="Delete versions with version numbers that match this version mask")
@click.option('--version', '-v', help='The version ID or the version-number')
@click.option('--format', '-f', default='{id} {versionNumber} {label} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False)
@pyke.auth.profile_context_required
def delete(widget, version_mask, version, profile, format, json):
    behavior.WidgetVersionBehavior(profile=profile, json=json).delete(widget, version, version_mask, format)

widget_version.add_command(list)
widget_version.add_command(create)
widget_version.add_command(start)
widget_version.add_command(stop)
widget_version.add_command(delete)
widget_version.add_command(update)
widget_version.add_command(run)
widget_version.add_command(logs)
